from django.conf import settings
from django.shortcuts import resolve_url
from django.urls import reverse
from django.utils.encoding import force_str


def get_app_title():
    """
    Title of application to display in IDP view
    """
    return force_str(
        getattr(settings, "DIGID_MOCK_APP_TITLE", "") or "DigiD Mock Login"
    )


def get_success_url():
    """
    Default URL to redirect to after login
    """
    url = (
        getattr(settings, "DIGID_MOCK_RETURN_URL", "")
        or getattr(settings, "LOGIN_REDIRECT_URL", "")
        or "/"
    )
    return resolve_url(url)


def get_cancel_url():
    """
    Default URL if user presses 'vorige/back' in IDP view
    """
    url = (
        getattr(settings, "DIGID_MOCK_CANCEL_URL", "")
        or getattr(settings, "LOGIN_URL", "")
        or "/"
    )
    return resolve_url(url)


def get_idp_login_url():
    """
    URL of the IDP login page (possibly on other domain)
    """
    url = getattr(settings, "DIGID_MOCK_IDP_LOGIN_URL", "") or reverse(
        "digid-mock:login"
    )
    return resolve_url(url)



# Generated by Django 3.2.18 on 2023-06-14 15:49

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("digid_eherkenning", "0002_auto_20221102_1046"),
    ]

    operations = [
        migrations.AddField(
            model_name="digidconfiguration",
            name="artifact_resolve_content_type",
            field=models.CharField(
                choices=[
                    ("application/soap+xml", "application/soap+xml"),
                    ("text/xml", "text/xml"),
                ],
                default="application/soap+xml",
                help_text="'application/soap+xml' is considered legacy and modern brokers typically expect 'text/xml'.",
                max_length=100,
                verbose_name="resolve artifact binding content type",
            ),
        ),
    ]



from django.urls import path

from . import views

app_name = "digid"

urlpatterns = [
    path("login/", views.DigiDLoginView.as_view(), name="login"),
    path("acs/", views.DigiDAssertionConsumerServiceView.as_view(), name="acs"),
    path("logout/", views.DigiDLogoutView.as_view(), name="logout"),
    path(
        "slo/redirect/",
        views.DigidSingleLogoutRedirectView.as_view(),
        name="slo-redirect",
    ),
    path("slo/soap/", views.DigidSingleLogoutSoapView.as_view(), name="slo-soap"),
]



from django.urls import path

from digid_eherkenning.mock.idp.views.digid import (
    DigiDMockIDPLoginView,
    DigiDMockIDPPasswordLoginView,
)

app_name = "digid-mock"

urlpatterns = [
    path("inloggen/", DigiDMockIDPLoginView.as_view(), name="login"),
    path("inloggen_ww/", DigiDMockIDPPasswordLoginView.as_view(), name="password"),
]



import urllib
from base64 import b64decode
from unittest.mock import patch

from django.conf import settings
from django.test import RequestFactory, TestCase, override_settings
from django.urls import reverse
from django.utils import timezone

import pytest
import responses
from freezegun import freeze_time
from furl import furl
from lxml import etree
from onelogin.saml2.utils import OneLogin_Saml2_Utils

from digid_eherkenning.choices import AssuranceLevels
from digid_eherkenning.models import EherkenningConfiguration
from digid_eherkenning.views import eHerkenningLoginView

from .project.models import User
from .utils import create_example_artifact, get_saml_element


@pytest.mark.usefixtures("eherkenning_config", "temp_private_root")
class eHerkenningLoginViewTests(TestCase):
    maxDiff = None

    @freeze_time("2020-04-09T08:31:46Z")
    @patch("onelogin.saml2.utils.uuid4")
    def test_login(self, uuid_mock):
        uuid_mock.hex = "80dd245883b84bd98dacbf3978af3d03"
        response = self.client.get(reverse("eherkenning:login"))

        saml_request = b64decode(
            response.context["form"].initial["SAMLRequest"].encode("utf-8")
        )

        tree = etree.fromstring(saml_request)

        self.assertEqual(
            tree.attrib,
            {
                "ID": "ONELOGIN_5ba93c9db0cff93f52b521d7420e43f6eda2784f",
                "Version": "2.0",
                "ForceAuthn": "true",
                "IssueInstant": "2020-04-09T08:31:46Z",
                "Destination": "https://eh01.staging.iwelcome.nl/broker/sso/1.13",
                "ProtocolBinding": "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Artifact",
                "AssertionConsumerServiceURL": "https://example.com/eherkenning/acs/",
            },
        )

        auth_context_class_ref = tree.xpath(
            "samlp:RequestedAuthnContext[@Comparison='minimum']/saml:AuthnContextClassRef",
            namespaces={
                "samlp": "urn:oasis:names:tc:SAML:2.0:protocol",
                "saml": "urn:oasis:names:tc:SAML:2.0:assertion",
            },
        )[0]

        self.assertEqual(
            auth_context_class_ref.text,
            "urn:etoegang:core:assurance-class:loa3",
        )

        # Make sure Signature properties are as expected.
        signature = tree.xpath(
            "//xmldsig:Signature",
            namespaces={"xmldsig": "http://www.w3.org/2000/09/xmldsig#"},
        )[0]

        elements = signature.xpath(
            "//xmldsig:SignatureValue",
            namespaces={"xmldsig": "http://www.w3.org/2000/09/xmldsig#"},
        )
        elements[0].text = ""

        elements = signature.xpath(
            "//xmldsig:DigestValue",
            namespaces={"xmldsig": "http://www.w3.org/2000/09/xmldsig#"},
        )
        elements[0].text = ""

        elements = signature.xpath(
            "//xmldsig:X509Certificate",
            namespaces={"xmldsig": "http://www.w3.org/2000/09/xmldsig#"},
        )
        elements[0].text = ""

        expected_signature = (
            '<ds:Signature xmlns:ds="http://www.w3.org/2000/09/xmldsig#" '
            ' xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol" '
            ' xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion">'
            "<ds:SignedInfo>"
            '<ds:CanonicalizationMethod Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#"/>'
            '<ds:SignatureMethod Algorithm="http://www.w3.org/2001/04/xmldsig-more#rsa-sha256"/>'
            '<ds:Reference URI="#ONELOGIN_5ba93c9db0cff93f52b521d7420e43f6eda2784f">'
            "<ds:Transforms>"
            '<ds:Transform Algorithm="http://www.w3.org/2000/09/xmldsig#enveloped-signature"/>'
            '<ds:Transform Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#"/>'
            "</ds:Transforms>"
            '<ds:DigestMethod Algorithm="http://www.w3.org/2001/04/xmlenc#sha256"/>'
            "<ds:DigestValue></ds:DigestValue>"
            "</ds:Reference>"
            "</ds:SignedInfo>"
            "<ds:SignatureValue></ds:SignatureValue>"
            "<ds:KeyInfo>"
            "<ds:X509Data>"
            "<ds:X509Certificate></ds:X509Certificate>"
            "</ds:X509Data>"
            "</ds:KeyInfo>"
            "</ds:Signature>"
        )

        self.assertXMLEqual(
            etree.tostring(signature, pretty_print=True).decode("utf-8"),
            etree.tostring(
                etree.fromstring(expected_signature), pretty_print=True
            ).decode("utf-8"),
        )

    def test_login_views_can_override_minimum_loa(self):
        class CustomLoginView(eHerkenningLoginView):
            def get_level_of_assurance(self):
                return (
                    AssuranceLevels.high
                    if "special" in self.request.GET.get("next")
                    else AssuranceLevels.middle
                )

        request = RequestFactory().get(reverse("eherkenning:login") + "?next=/special")

        response = CustomLoginView.as_view()(request)

        saml_request = b64decode(
            response.context_data["form"].initial["SAMLRequest"].encode("utf-8")
        )
        tree = etree.fromstring(saml_request)
        auth_context_class_ref = tree.xpath(
            "samlp:RequestedAuthnContext[@Comparison='minimum']/saml:AuthnContextClassRef",
            namespaces={
                "samlp": "urn:oasis:names:tc:SAML:2.0:protocol",
                "saml": "urn:oasis:names:tc:SAML:2.0:assertion",
            },
        )[0]

        self.assertEqual(auth_context_class_ref.text, AssuranceLevels.high.value)

    @freeze_time("2020-04-09T08:31:46Z")
    @patch("onelogin.saml2.utils.uuid4")
    def test_login_with_attribute_consuming_service_index(self, uuid_mock):
        uuid_mock.hex = "80dd245883b84bd98dacbf3978af3d03"
        url = furl(reverse("eherkenning:login")).set(
            {"attr_consuming_service_index": "2"}
        )

        response = self.client.get(url)

        saml_request = b64decode(
            response.context["form"].initial["SAMLRequest"].encode("utf-8")
        )

        tree = etree.fromstring(saml_request)

        self.assertEqual(
            tree.attrib,
            {
                "ID": "ONELOGIN_5ba93c9db0cff93f52b521d7420e43f6eda2784f",
                "Version": "2.0",
                "ForceAuthn": "true",
                "IssueInstant": "2020-04-09T08:31:46Z",
                "Destination": "https://eh01.staging.iwelcome.nl/broker/sso/1.13",
                "ProtocolBinding": "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Artifact",
                "AssertionConsumerServiceURL": "https://example.com/eherkenning/acs/",
                "AttributeConsumingServiceIndex": "2",
            },
        )


@pytest.mark.usefixtures("eherkenning_config", "temp_private_root")
@freeze_time("2020-04-09T08:31:46Z")
class eHerkenningAssertionConsumerServiceViewTests(TestCase):
    def setUp(self):
        super().setUp()

        config = EherkenningConfiguration.get_solo()

        with config.certificate.public_certificate.open("r") as cert_file:
            cert = cert_file.read()

        encrypted_attribute = OneLogin_Saml2_Utils.generate_name_id(
            "123456782",
            sp_nq=None,
            nq="urn:etoegang:1.9:EntityConcernedID:RSIN",
            sp_format="urn:oasis:names:tc:SAML:2.0:nameid-format:persistent",
            cert=cert,
        )

        self.bogus_signature = (
            "<ds:Signature>"
            "<ds:SignedInfo>"
            '<ds:CanonicalizationMethod Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#"/>'
            '<ds:SignatureMethod Algorithm="http://www.w3.org/2001/04/xmldsig-more#rsa-sha256"/>'
            '<ds:Reference URI="#{id}">'
            "<ds:Transforms>"
            '<ds:Transform Algorithm="http://www.w3.org/2000/09/xmldsig#enveloped-signature"/>'
            '<ds:Transform Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#">'
            '<ec:InclusiveNamespaces xmlns:ec="http://www.w3.org/2001/10/xml-exc-c14n#" PrefixList="xacml-saml"/>'
            "</ds:Transform>"
            "</ds:Transforms>"
            '<ds:DigestMethod Algorithm="http://www.w3.org/2001/04/xmlenc#sha256"/>'
            "<ds:DigestValue></ds:DigestValue>"
            "</ds:Reference>"
            "</ds:SignedInfo>"
            "<ds:SignatureValue>"
            ""
            "</ds:SignatureValue>"
            "<ds:KeyInfo>"
            "<ds:KeyName></ds:KeyName>"
            "</ds:KeyInfo>"
            "</ds:Signature>"
        )
        # self.bogus_signature = (
        #     '<ds:Signature xmlns:ds="http://www.w3.org/2000/09/xmldsig#">'
        #     '<ds:SignedInfo>'
        #     '<ds:CanonicalizationMethod Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#"/>'
        #     '<ds:SignatureMethod Algorithm="http://www.w3.org/2001/04/xmldsig-more#rsa-sha256"/>'
        #     '<ds:Reference URI="#_0ddd4451-264c-3823-88e2-7da7490652cd">'
        #     '<ds:Transforms>'
        #     '<ds:Transform Algorithm="http://www.w3.org/2000/09/xmldsig#enveloped-signature"/>'
        #     '<ds:Transform Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#">'
        #     '<ec:InclusiveNamespaces xmlns:ec="http://www.w3.org/2001/10/xml-exc-c14n#" PrefixList="xacml-saml"/>'
        #     '</ds:Transform>'
        #     '</ds:Transforms>'
        #     '<ds:DigestMethod Algorithm="http://www.w3.org/2001/04/xmlenc#sha256"/>'
        #     '<ds:DigestValue>/q9QSh5W8fF0+UxcvJ3tbxPGSD4d66BGekaHyqH+oX4=</ds:DigestValue>'
        #     '</ds:Reference>'
        #     '</ds:SignedInfo>'
        #     '<ds:SignatureValue>'
        #     'zxgJZd8Y1w/xce3ptMqSOLlA7MSv9r7hG0x+XQYQohSJpldEp5/ZVV6TyxonMvmKSJxl7KNoMvK9'
        #     'XrXDuy02L0oIchUFIUfU5O1h5IouON6WRuQjhcILlL/hhWayqwabiDJ8iqAoifgmSRM1A/Am0+6c'
        #     '9oTULCLjk3OtHZXcXb0VWJGM9CHvLiG2rWJtggxhJOFX0TQ5AUIkDtilN74flQSyH5bAlXSnkyo5'
        #     'Z77nQ4NcdWctpOSgnwx5fFHg69IWac8DjYs2/eQ72AIDsoEgb7x/qtCchseJSbm6rCDJWi8qzMDj'
        #     '0uw0mnxf1OrrLq2Mmz5hopGn0y+ueGwCDsNwY2Bd1DgifqzH8ra5asI63rkPghOuM7x96Ovob2lx'
        #     'bJAXVkZXinIsCxVrNTSPXIjQiLs+uHkM/rDa31a9XXGRddTekOI449ZRxgvlMcp2SViIPmBWv8Fe'
        #     'rbgriNaRZ2Kr2oa1sXcc02UGwDvJ6jX+q2EXd38txiuW254LzI9P9FenW7CQsuKR9ArIW9XWyQnI'
        #     'FB9X/mWKZXxVsf8yhlQ9mgDb3xtvQ326TYD9PuCVInRmsBVATVGJs64qEEaJq17XaL52JzXZicK/'
        #     'rb8ciC3U/vruE5OWcsORQEivG09LcDu9cFhFLjSuPtaEbAS34rVKIsmNLJvbg3e/qaS2oMszEP4='
        #     '</ds:SignatureValue>'
        #     '<ds:KeyInfo>'
        #     '<ds:KeyName>e6e04e0a22bbc8a036a8a243abc9655e92907f73a4ba5a2ad28485ec3f4c82d1</ds:KeyName>'
        #     '</ds:KeyInfo>'
        #     '</ds:Signature>'
        # )

        # eHerkenning has a Advice element with more elements than this. But these elements are what
        # broke python3-saml and for which I had to introduce "disableSignatureWrappingProtection" security setting.
        self.advice = (
            "<saml:Advice>"
            '<saml:Assertion ID="bla" IssueInstant="2020-04-09T08:31:46Z" Version="2.0">'
            "<saml:Issuer>urn:etoegang:HM:00000003520354760000:entities:9632</saml:Issuer>"
            + self.bogus_signature.format(id="bla")
            + "</saml:Assertion>"
            "</saml:Advice>"
        )
        self.assertion = (
            '<saml:Assertion ID="_ae28e39f-bf7a-32d5-9653-3ad07c0e911e" IssueInstant="2020-04-09T08:31:46Z" Version="2.0" xmlns:xacml-saml="urn:oasis:xacml:2.0:saml:assertion:schema:os">'
            "<saml:Issuer>urn:etoegang:HM:00000003520354760000:entities:9632</saml:Issuer>"
            + self.bogus_signature.format(id="_ae28e39f-bf7a-32d5-9653-3ad07c0e911e")
            + "<saml:Subject>"
            '<saml:NameID Format="urn:oasis:names:tc:SAML:2.0:nameid-format:transient" NameQualifier="urn:etoegang:EB:00000004000000149000:entities:9009">b964780b-3441-4e57-a027-a59c21c3019d</saml:NameID>'
            '<saml:SubjectConfirmation Method="urn:oasis:names:tc:SAML:2.0:cm:bearer">'
            '<saml:SubjectConfirmationData InResponseTo="id-jiaDzLL9mR3C3hioH" NotOnOrAfter="2020-04-09T08:35:46Z" Recipient="https://example.com/eherkenning/acs/"/>'
            "</saml:SubjectConfirmation>"
            "</saml:Subject>"
            '<saml:Conditions NotBefore="2020-04-09T08:31:46Z" NotOnOrAfter="2020-04-09T08:35:46Z">'
            "<saml:AudienceRestriction>"
            "<saml:Audience>urn:etoegang:DV:0000000000000000001:entities:0002</saml:Audience>"
            "</saml:AudienceRestriction>"
            "</saml:Conditions>"
            + self.advice
            + '<saml:AuthnStatement AuthnInstant="2020-05-06T10:50:14Z">'
            "<saml:AuthnContext>"
            "<saml:AuthnContextClassRef>urn:etoegang:core:assurance-class:loa3</saml:AuthnContextClassRef>"
            "<saml:AuthenticatingAuthority>urn:etoegang:EB:00000004000000149000:entities:9009</saml:AuthenticatingAuthority>"
            "</saml:AuthnContext>"
            "</saml:AuthnStatement>"
            "<saml:AttributeStatement>"
            '<saml:Attribute Name="urn:etoegang:core:ServiceID">'
            '<saml:AttributeValue xsi:type="xs:string">urn:etoegang:DV:00000002003214394001:services:5000</saml:AttributeValue>'
            "</saml:Attribute>"
            '<saml:Attribute Name="urn:etoegang:core:ServiceUUID">'
            '<saml:AttributeValue xsi:type="xs:string">87f3035b-b0c2-482a-b693-98316f5f4ba4</saml:AttributeValue>'
            "</saml:Attribute>"
            '<saml:Attribute FriendlyName="ActingSubjectID" Name="urn:etoegang:core:LegalSubjectID">'
            "<saml:AttributeValue>"
            + encrypted_attribute
            + "</saml:AttributeValue></saml:Attribute>"
            "</saml:AttributeStatement>"
            "</saml:Assertion>"
        )

        self.response = (
            "<samlp:Response"
            ' Destination="https://example.com/eherkenning/acs/"'
            ' ID="_d4d73890-b5ca-3ca4-ab7b-d078378e3527"'
            ' InResponseTo="id-jiaDzLL9mR3C3hioH"'
            ' IssueInstant="2020-04-09T08:31:46Z"'
            ' Version="2.0"'
            ' xmlns:ds="http://www.w3.org/2000/09/xmldsig#"'
            ' xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion"'
            ' xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol"'
            ' xmlns:xs="http://www.w3.org/2001/XMLSchema"'
            ' xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
            "<saml:Issuer>urn:etoegang:HM:00000003520354760000:entities:9632</saml:Issuer>"
            "<samlp:Status>"
            '<samlp:StatusCode Value="urn:oasis:names:tc:SAML:2.0:status:Success"/>'
            "</samlp:Status>" + self.assertion + "</samlp:Response>"
        )
        self.artifact_response = (
            "<samlp:ArtifactResponse"
            ' xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol"'
            ' xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion"'
            ' xmlns:ds="http://www.w3.org/2000/09/xmldsig#"'
            ' xmlns:ec="http://www.w3.org/2001/10/xml-exc-c14n#"'
            ' ID="_1330416516" Version="2.0" IssueInstant="2020-04-09T08:31:46Z"'
            ' InResponseTo="ONELOGIN_5ba93c9db0cff93f52b521d7420e43f6eda2784f">'
            "<saml:Issuer>urn:etoegang:HM:00000003520354760000:entities:9632</saml:Issuer>"
            "<samlp:Status>"
            '<samlp:StatusCode Value="urn:oasis:names:tc:SAML:2.0:status:Success"/>'
            "</samlp:Status>" + self.response + "</samlp:ArtifactResponse>"
        )

        self.artifact_response_soap = (
            b'<?xml version="1.0" encoding="UTF-8"?>'
            b"<soapenv:Envelope"
            b' xmlns:soapenv="http://schemas.xmlsoap.org/soap/envelope/"'
            b' xmlns:xsd="http://www.w3.org/2001/XMLSchema"'
            b' xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
            b"<soapenv:Body>"
            + str(self.artifact_response).encode("utf-8")
            + b"</soapenv:Body>"
            b"</soapenv:Envelope>"
        )

        self.artifact = create_example_artifact(
            "urn:etoegang:HM:00000003520354760000:entities:9632",
            endpoint_index=b"\x00\x01",
        )

        self.uuid_patcher = patch("onelogin.saml2.utils.uuid4")
        self.cache_patcher = patch("digid_eherkenning.saml2.base.cache")

        self.uuid_mock = self.uuid_patcher.start()
        self.uuid_mock.hex = "80dd245883b84bd98dacbf3978af3d03"

        self.cache_mock = self.cache_patcher.start()
        self.cache_mock.get.return_value = {
            "current_time": timezone.now(),
            "client_ip_address": "127.0.0.1",
        }

        self.validate_sign_patcher = patch.object(OneLogin_Saml2_Utils, "validate_sign")
        self.validate_sign_mock = self.validate_sign_patcher.start()

        self.addCleanup(patch.stopall)

    @responses.activate
    def test_get(self):
        responses.add(
            responses.POST,
            "https://eh02.staging.iwelcome.nl/broker/ars/1.13",
            body=self.artifact_response_soap,
            status=200,
        )
        url = (
            reverse("eherkenning:acs")
            + "?"
            + urllib.parse.urlencode({"SAMLart": self.artifact, "RelayState": "/home/"})
        )

        with self.assertLogs("digid_eherkenning.backends", level="INFO") as log_watcher:
            response = self.client.get(url, secure=True)

        logs = [r.getMessage() for r in log_watcher.records]
        self.assertIn(
            "User user-123456782 (new account) from 127.0.0.1 logged in using eHerkenning",
            logs,
        )

        # Make sure we're redirect the the right place.
        self.assertEqual(response.url, "/home/")

        # Make sure that the cache is checked for the InResponseTo returned
        # by the IDP.
        self.cache_mock.get.assert_called_once_with("eherkenning_id-jiaDzLL9mR3C3hioH")

    @override_settings(LOGIN_URL="/dummy/login")
    @responses.activate
    def test_no_authn_request(self):
        """
        Make sure that when the InResponseTo in the Response does not match
        any id we've given out, an error occurs.
        """
        self.cache_mock.get.return_value = None

        responses.add(
            responses.POST,
            "https://eh02.staging.iwelcome.nl/broker/ars/1.13",
            body=self.artifact_response_soap,
            status=200,
        )
        url = (
            reverse("eherkenning:acs")
            + "?"
            + urllib.parse.urlencode({"SAMLart": self.artifact})
        )

        with self.assertLogs("digid_eherkenning.backends", level="INFO") as log_watcher:
            response = self.client.get(url)

        logs = [r.getMessage() for r in log_watcher.records]

        self.assertIn(
            "A technical error occurred from 127.0.0.1 during eHerkenning login.", logs
        )

        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/dummy/login")
        # Make sure no user is created.
        self.assertEqual(User.objects.count(), 0)

    @responses.activate
    def test_redirect_default(self):
        """
        Make sure the view returns to the default URL if no RelayState is set
        """
        responses.add(
            responses.POST,
            "https://eh02.staging.iwelcome.nl/broker/ars/1.13",
            body=self.artifact_response_soap,
            status=200,
        )
        url = (
            reverse("eherkenning:acs")
            + "?"
            + urllib.parse.urlencode({"SAMLart": self.artifact})
        )
        response = self.client.get(url)

        self.assertEqual(response.url, settings.LOGIN_REDIRECT_URL)

    # TODO: Add authnfailed tests here as well.

    @override_settings(LOGIN_URL=reverse("admin:login"))
    @responses.activate
    def test_no_rsin(self):
        artifact_response_soap = etree.fromstring(self.artifact_response_soap)

        # Remove the RSIN. In this scenario it is not returned by eHerkenning.
        encrypted_id = get_saml_element(
            artifact_response_soap,
            "//saml:EncryptedID",
        )
        encrypted_id.getparent().remove(encrypted_id)

        responses.add(
            responses.POST,
            "https://eh02.staging.iwelcome.nl/broker/ars/1.13",
            body=etree.tostring(artifact_response_soap),
            status=200,
        )

        url = (
            reverse("eherkenning:acs")
            + "?"
            + urllib.parse.urlencode({"SAMLart": self.artifact})
        )

        response = self.client.get(url, follow=True)

        messages = [str(m) for m in response.context["messages"]]
        self.assertIn(
            "No RSIN returned by eHerkenning. Login to eHerkenning did not succeed.",
            messages,
        )

    @responses.activate
    def test_user_cancels(self):
        """
        Test that when a user cancels this is logged properly.
        """

        artifact_response_soap = etree.fromstring(self.artifact_response_soap)

        # Remove Assertion element. It will not be returned
        # when user cancels.
        assertion = get_saml_element(
            artifact_response_soap,
            "//samlp:Response/saml:Assertion",
        )
        assertion.getparent().remove(assertion)

        status_code = get_saml_element(
            artifact_response_soap, "//samlp:Response/samlp:Status/samlp:StatusCode"
        )
        status_code.set("Value", "urn:oasis:names:tc:SAML:2.0:status:Responder")

        status_code.insert(
            0,
            etree.Element(
                "{urn:oasis:names:tc:SAML:2.0:protocol}StatusCode",
                Value="urn:oasis:names:tc:SAML:2.0:status:AuthnFailed",
            ),
        )

        responses.add(
            responses.POST,
            "https://eh02.staging.iwelcome.nl/broker/ars/1.13",
            body=etree.tostring(artifact_response_soap),
            status=200,
        )

        url = (
            reverse("eherkenning:acs")
            + "?"
            + urllib.parse.urlencode({"SAMLart": self.artifact})
        )

        with self.assertLogs("digid_eherkenning.backends", level="INFO") as log_watcher:
            response = self.client.get(url)

        logs = [r.getMessage() for r in log_watcher.records]

        self.assertIn(
            "The eHerkenning login from 127.0.0.1 did not succeed or was cancelled.",
            logs,
        )

        self.assertEqual(response.status_code, 302)
        # Make sure no user is created.
        self.assertEqual(User.objects.count(), 0)



import urllib
from base64 import b64decode
from unittest import skip
from unittest.mock import patch

from django.conf import settings
from django.contrib import auth
from django.test import RequestFactory, TestCase, override_settings
from django.urls import reverse
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

import pytest
import responses
from freezegun import freeze_time
from lxml import etree
from onelogin.saml2.utils import OneLogin_Saml2_Utils

from digid_eherkenning.choices import DigiDAssuranceLevels
from digid_eherkenning.views import DigiDLoginView

from .project.models import User
from .utils import create_example_artifact, get_saml_element


@pytest.mark.usefixtures("digid_config", "temp_private_root")
class DigidLoginViewTests(TestCase):
    maxDiff = None

    @freeze_time("2020-04-09T08:31:46Z")
    @patch("onelogin.saml2.utils.uuid4")
    def test_login(self, uuid_mock):
        """
        DigID

        Make sure DigiD - 3.3.2 Stap 2 Authenticatievraag

        works as intended.
        """
        uuid_mock.hex = "80dd245883b84bd98dacbf3978af3d03"
        response = self.client.get(reverse("digid:login"))

        saml_request = b64decode(
            response.context["form"].initial["SAMLRequest"].encode("utf-8")
        )

        #
        # DigiD - 1.4 Voorbeeldbericht bij Stap 2 : AuthnRequest Post Binding
        #
        # <?xml version="1.0" encoding="UTF-8"?>
        # <samlp:AuthnRequest
        #  xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol"
        #  xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion"
        #  xmlns:ds="http://www.w3.org/2000/09/xmldsig#"
        #  xmlns:ec="http://www.w3.org/2001/10/xml-exc-c14n#"
        #  Destination="https://example.com" ForceAuthn="false" ID="_1330416073" Version="2.0"
        #  IssueInstant="2012-02-28T09:01:13Z" AssertionConsumerServiceIndex="0"
        #  ProviderName="provider name">
        #    <saml:Issuer>https://sp.example.com</saml:Issuer>
        #    <ds:Signature><!—Zie XML Signature--></ds:Signature>
        #    <samlp:RequestedAuthnContext Comparison="minimum">
        #      <saml:AuthnContextClassRef>
        #      urn:oasis:names:tc:SAML:2.0:ac:classes:PasswordProtectedTransport
        #      </saml:AuthnContextClassRef>
        #    </samlp:RequestedAuthnContext>
        # </samlp:AuthnRequest>
        #

        # DigiD - 1.1 Xml Signature
        # <ds:Signature>
        #  <ds:SignedInfo>
        #  <ds:CanonicalizationMethod Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#"/>
        #  <ds:SignatureMethod Algorithm="http://www.w3.org/2001/04/xmldsig-more#rsa-sha256"/>
        #  <ds:Reference URI="#_1330416073">
        #  <ds:Transforms>
        #  <ds:Transform Algorithm="http://www.w3.org/2000/09/xmldsig#enveloped-signature"/>
        #  <ds:Transform Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#">
        #  <ec:InclusiveNamespaces PrefixList="ds saml samlp xs"/>
        #  </ds:Transform>
        #  </ds:Transforms>
        #  <ds:DigestMethod Algorithm="http://www.w3.org/2001/04/xmlenc#sha256"/>
        #  <ds:DigestValue>irsh4GNXQcsbkUmex22XsUejBTXyDdHfaUL/MFFWQHs=</ds:DigestValue>
        #  </ds:Reference>
        #  </ds:SignedInfo>
        #  <ds:SignatureValue>YJ0V4gCTwRYvgy <INGEKORT> LnOEvyF2ddwBFwILL4nCpw==</ds:SignatureValue>
        # </ds:Signature>

        tree = etree.fromstring(saml_request)

        self.assertEqual(
            tree.attrib,
            {
                "ID": "ONELOGIN_5ba93c9db0cff93f52b521d7420e43f6eda2784f",
                "Version": "2.0",
                "IssueInstant": "2020-04-09T08:31:46Z",
                "Destination": "https://preprod1.digid.nl/saml/idp/request_authentication",
                "ProtocolBinding": "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Artifact",
                "AssertionConsumerServiceURL": "https://sp.example.nl/digid/acs/",
            },
        )

        auth_context_class_ref = tree.xpath(
            "samlp:RequestedAuthnContext[@Comparison='minimum']/saml:AuthnContextClassRef",
            namespaces={
                "samlp": "urn:oasis:names:tc:SAML:2.0:protocol",
                "saml": "urn:oasis:names:tc:SAML:2.0:assertion",
            },
        )[0]

        self.assertEqual(
            auth_context_class_ref.text,
            "urn:oasis:names:tc:SAML:2.0:ac:classes:MobileTwoFactorContract",
        )

        # Make sure Signature properties are as expected.
        signature = tree.xpath(
            "//xmldsig:Signature",
            namespaces={"xmldsig": "http://www.w3.org/2000/09/xmldsig#"},
        )[0]

        elements = signature.xpath(
            "//xmldsig:SignatureValue",
            namespaces={"xmldsig": "http://www.w3.org/2000/09/xmldsig#"},
        )
        elements[0].text = ""

        elements = signature.xpath(
            "//xmldsig:DigestValue",
            namespaces={"xmldsig": "http://www.w3.org/2000/09/xmldsig#"},
        )
        elements[0].text = ""

        elements = signature.xpath(
            "//xmldsig:X509Certificate",
            namespaces={"xmldsig": "http://www.w3.org/2000/09/xmldsig#"},
        )
        elements[0].text = ""

        expected_signature = (
            '<ds:Signature xmlns:ds="http://www.w3.org/2000/09/xmldsig#" '
            ' xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol" '
            ' xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion">'
            "<ds:SignedInfo>"
            '<ds:CanonicalizationMethod Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#"/>'
            '<ds:SignatureMethod Algorithm="http://www.w3.org/2001/04/xmldsig-more#rsa-sha256"/>'
            '<ds:Reference URI="#ONELOGIN_5ba93c9db0cff93f52b521d7420e43f6eda2784f">'
            "<ds:Transforms>"
            '<ds:Transform Algorithm="http://www.w3.org/2000/09/xmldsig#enveloped-signature"/>'
            '<ds:Transform Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#"/>'
            "</ds:Transforms>"
            '<ds:DigestMethod Algorithm="http://www.w3.org/2001/04/xmlenc#sha256"/>'
            "<ds:DigestValue></ds:DigestValue>"
            "</ds:Reference>"
            "</ds:SignedInfo>"
            "<ds:SignatureValue></ds:SignatureValue>"
            "<ds:KeyInfo>"
            "<ds:X509Data>"
            "<ds:X509Certificate></ds:X509Certificate>"
            "</ds:X509Data>"
            "</ds:KeyInfo>"
            "</ds:Signature>"
        )

        self.assertXMLEqual(
            etree.tostring(signature, pretty_print=True).decode("utf-8"),
            etree.tostring(
                etree.fromstring(expected_signature), pretty_print=True
            ).decode("utf-8"),
        )

    def test_login_views_can_override_minimum_loa(self):
        class CustomLoginView(DigiDLoginView):
            def get_level_of_assurance(self):
                return (
                    DigiDAssuranceLevels.substantial
                    if "special" in self.request.GET.get("next")
                    else DigiDAssuranceLevels.middle
                )

        request = RequestFactory().get(reverse("digid:login") + "?next=/special")

        response = CustomLoginView.as_view()(request)

        saml_request = b64decode(
            response.context_data["form"].initial["SAMLRequest"].encode("utf-8")
        )
        tree = etree.fromstring(saml_request)
        auth_context_class_ref = tree.xpath(
            "samlp:RequestedAuthnContext[@Comparison='minimum']/saml:AuthnContextClassRef",
            namespaces={
                "samlp": "urn:oasis:names:tc:SAML:2.0:protocol",
                "saml": "urn:oasis:names:tc:SAML:2.0:assertion",
            },
        )[0]

        self.assertEqual(
            auth_context_class_ref.text, DigiDAssuranceLevels.substantial.value
        )


@freeze_time("2020-04-09T08:31:46Z")
@override_settings(LOGIN_URL=reverse("admin:login"))
@pytest.mark.usefixtures("digid_config", "temp_private_root")
class DigidAssertionConsumerServiceViewTests(TestCase):
    maxDiff = None

    def setUp(self):
        super().setUp()

        # DigiD - 1.6 Voorbeeldbericht bij Stap 7 : Artifact Response (SOAP)
        # In een Soap envelope. Voor de leesbaarheid is de Saml Assertion uit de Response genomen.

        # <samlp:ArtifactResponse
        #  xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol"
        #  xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion"
        #  xmlns:ds="http://www.w3.org/2000/09/xmldsig#"
        #  xmlns:ec="http://www.w3.org/2001/10/xml-exc-c14n#"
        #  ID="_1330416516" Version="2.0" IssueInstant="2012-12-20T18:50:27Z"
        #  InResponseTo="_1330416516">
        #  <saml:Issuer>https://idp.example.com</saml:Issuer>
        #  <ds:Signature><!-- Zie XML Signature --></ds:Signature>
        #  <samlp:Status>
        #  <samlp:StatusCode Value="urn:oasis:names:tc:SAML:2.0:status:Success"/>
        #  </samlp:Status>
        #  <samlp:Response InResponseTo="_7afa5ce49" Version="2.0" ID="_1072ee96"
        #  IssueInstant="2012-12-20T18:50:27Z">
        #  <saml:Issuer>https://idp.example.com</saml:Issuer>
        #  <samlp:Status>
        #  <samlp:StatusCode Value="urn:oasis:names:tc:SAML:2.0:status:Success"/>
        #  </samlp:Status>
        #  <saml:Assertion><!—ZIE ASSERTION HIERONDER --></saml:Assertion>
        #  </samlp:Response>
        # </samlp:ArtifactResponse>

        # <saml:Assertion Version="2.0" ID="_dc9f70e61c" IssueInstant="2012-12-20T18:50:27Z">
        #  <saml:Issuer>https://idp.example.com</saml:Issuer>
        #  <ds:Signature><!—Optioneel Zie XML Signature --></ds:Signature>
        #  <saml:Subject>
        #  <saml:NameID>s00000000:12345678</saml:NameID>
        #  <saml:SubjectConfirmation Method="urn:oasis:names:tc:SAML:2.0:cm:bearer">
        #  <saml:SubjectConfirmationData InResponseTo="_7afa5ce49"
        #  Recipient="http://example.com/artifact_url" NotOnOrAfter="2012-12-20T18:52:27Z"/>
        #  </saml:SubjectConfirmation>
        #  </saml:Subject>
        #  <saml:Conditions NotBefore="2012-12-20T18:48:27Z" NotOnOrAfter="2012-12-20T18:52:27Z">
        #  <saml:AudienceRestriction>
        #  <saml:Audience>http://sp.example.com</saml:Audience>
        #  </saml:AudienceRestriction>
        #  </saml:Conditions>
        #  <saml:AuthnStatement SessionIndex="17" AuthnInstant="2012-12-20T18:50:27Z">
        #  <saml:SubjectLocality Address="127.0.0.1"/>
        #  <saml:AuthnContext Comparison="minimum">
        #  <saml:AuthnContextClassRef>
        #  urn:oasis:names:tc:SAML:2.0:ac:classes:PasswordProtectedTransport
        #  </saml:AuthnContextClassRef>
        #  </saml:AuthnContext>
        #  </saml:AuthnStatement>
        # </saml:Assertion>

        self.bogus_signature = (
            "<ds:Signature>"
            "<ds:SignedInfo>"
            '<ds:CanonicalizationMethod Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#"/>'
            '<ds:SignatureMethod Algorithm="http://www.w3.org/2001/04/xmldsig-more#rsa-sha256"/>'
            '<ds:Reference URI="#{id}">'
            "<ds:Transforms>"
            '<ds:Transform Algorithm="http://www.w3.org/2000/09/xmldsig#enveloped-signature"/>'
            '<ds:Transform Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#">'
            '<ec:InclusiveNamespaces xmlns:ec="http://www.w3.org/2001/10/xml-exc-c14n#" PrefixList="xacml-saml"/>'
            "</ds:Transform>"
            "</ds:Transforms>"
            '<ds:DigestMethod Algorithm="http://www.w3.org/2001/04/xmlenc#sha256"/>'
            "<ds:DigestValue></ds:DigestValue>"
            "</ds:Reference>"
            "</ds:SignedInfo>"
            "<ds:SignatureValue>"
            ""
            "</ds:SignatureValue>"
            "<ds:KeyInfo>"
            "<ds:KeyName></ds:KeyName>"
            "</ds:KeyInfo>"
            "</ds:Signature>"
        )

        self.response = (
            '<samlp:Response InResponseTo="_7afa5ce49" Version="2.0" ID="_1072ee96"'
            ' IssueInstant="2020-04-09T08:31:46Z">'
            "<saml:Issuer>https://was-preprod1.digid.nl/saml/idp/metadata</saml:Issuer>"
            + self.bogus_signature.format(id="_1072ee96")
            + "<samlp:Status>"
            '<samlp:StatusCode Value="urn:oasis:names:tc:SAML:2.0:status:Success"/>'
            "</samlp:Status>"
            '<saml:Assertion Version="2.0" ID="_dc9f70e61c" IssueInstant="2020-04-09T08:31:46Z">'
            "<saml:Issuer>https://was-preprod1.digid.nl/saml/idp/metadata</saml:Issuer>"
            "<saml:Subject>"
            "<saml:NameID>s00000000:12345678</saml:NameID>"
            '<saml:SubjectConfirmation Method="urn:oasis:names:tc:SAML:2.0:cm:bearer">'
            '<saml:SubjectConfirmationData InResponseTo="_7afa5ce49"'
            ' Recipient="https://sp.example.nl/digid/acs/" NotOnOrAfter="2020-04-10T08:31:46Z"/>'
            "</saml:SubjectConfirmation>"
            "</saml:Subject>"
            '<saml:Conditions NotBefore="2012-12-20T18:48:27Z" NotOnOrAfter="2020-04-10T08:31:46Z">'
            "<saml:AudienceRestriction>"
            "<saml:Audience>sp.example.nl/digid</saml:Audience>"
            "</saml:AudienceRestriction>"
            "</saml:Conditions>"
            '<saml:AuthnStatement SessionIndex="17" AuthnInstant="2020-04-09T08:31:46Z">'
            '<saml:SubjectLocality Address="127.0.0.1"/>'
            "<saml:AuthnContext>"
            "<saml:AuthnContextClassRef>"
            " urn:oasis:names:tc:SAML:2.0:ac:classes:PasswordProtectedTransport"
            "</saml:AuthnContextClassRef>"
            "</saml:AuthnContext>"
            "</saml:AuthnStatement>"
            "</saml:Assertion>"
            "</samlp:Response>"
        )

        self.artifact_response = (
            "<samlp:ArtifactResponse"
            ' xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol"'
            ' xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion"'
            ' xmlns:ds="http://www.w3.org/2000/09/xmldsig#"'
            ' xmlns:ec="http://www.w3.org/2001/10/xml-exc-c14n#"'
            ' ID="_1330416516" Version="2.0" IssueInstant="2020-04-09T08:31:46Z"'
            ' InResponseTo="ONELOGIN_5ba93c9db0cff93f52b521d7420e43f6eda2784f">'
            "<saml:Issuer>https://was-preprod1.digid.nl/saml/idp/metadata</saml:Issuer>"
            "<samlp:Status>"
            '<samlp:StatusCode Value="urn:oasis:names:tc:SAML:2.0:status:Success"/>'
            "</samlp:Status>" + self.response + "</samlp:ArtifactResponse>"
        )

        self.artifact_response_soap = (
            b'<?xml version="1.0" encoding="UTF-8"?>'
            b"<soapenv:Envelope"
            b' xmlns:soapenv="http://schemas.xmlsoap.org/soap/envelope/"'
            b' xmlns:xsd="http://www.w3.org/2001/XMLSchema"'
            b' xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
            b"<soapenv:Body>"
            + str(self.artifact_response).encode("utf-8")
            + b"</soapenv:Body>"
            b"</soapenv:Envelope>"
        )

        self.artifact = create_example_artifact(
            "https://was-preprod1.digid.nl/saml/idp/metadata"
        )
        self.uuid_patcher = patch("onelogin.saml2.utils.uuid4")
        self.cache_patcher = patch("digid_eherkenning.saml2.base.cache")

        self.uuid_mock = self.uuid_patcher.start()
        self.uuid_mock.hex = "80dd245883b84bd98dacbf3978af3d03"

        self.cache_mock = self.cache_patcher.start()
        self.cache_mock.get.return_value = {
            "current_time": timezone.now(),
            "client_ip_address": "127.0.0.1",
        }

        self.validate_sign_patcher = patch.object(OneLogin_Saml2_Utils, "validate_sign")
        self.validate_sign_mock = self.validate_sign_patcher.start()

        self.addCleanup(patch.stopall)

    @responses.activate
    def test_response_status_code_authnfailed(self):
        root_element = etree.fromstring(self.artifact_response_soap)

        # Remove Assertion element. It will not be returned
        # when user cancels.
        assertion = get_saml_element(
            root_element,
            "//saml:Assertion",
        )
        assertion.getparent().remove(assertion)

        status_code = get_saml_element(
            root_element, "//samlp:Response/samlp:Status/samlp:StatusCode"
        )
        status_code.set("Value", "urn:oasis:names:tc:SAML:2.0:status:Responder")

        status_code.insert(
            0,
            etree.Element(
                "{urn:oasis:names:tc:SAML:2.0:protocol}StatusCode",
                Value="urn:oasis:names:tc:SAML:2.0:status:NoAvailableIDP",
            ),
        )

        responses.add(
            responses.POST,
            "https://was-preprod1.digid.nl/saml/idp/resolve_artifact",
            body=etree.tostring(root_element),
            status=200,
        )

        url = (
            reverse("digid:acs")
            + "?"
            + urllib.parse.urlencode({"SAMLart": self.artifact})
        )
        with self.assertLogs("digid_eherkenning.backends", level="INFO") as log_watcher:
            response = self.client.get(url, follow=True)

        logs = [r.getMessage() for r in log_watcher.records]
        self.assertIn(
            "A technical error occurred from 127.0.0.1 during DigiD login.", logs
        )

        self.assertEqual(response.redirect_chain, [("/admin/login/", 302)])
        self.assertEqual(
            list(response.context["messages"])[0].message,
            _(
                "An error occurred in the communication with DigiD. "
                "Please try again later. If this error persists, please "
                "check the website https://www.digid.nl for the latest information."
            ),
        )

        # Make sure no user is created.
        self.assertEqual(User.objects.count(), 0)

    @responses.activate
    def test_artifact_response_status_code_authnfailed(self):
        root_element = etree.fromstring(self.artifact_response_soap)
        status_code = get_saml_element(
            root_element, "//samlp:ArtifactResponse/samlp:Status/samlp:StatusCode"
        )
        status_code.set("Value", "urn:oasis:names:tc:SAML:2.0:status:AuthnFailed")

        responses.add(
            responses.POST,
            "https://was-preprod1.digid.nl/saml/idp/resolve_artifact",
            body=etree.tostring(root_element),
            status=200,
        )

        url = (
            reverse("digid:acs")
            + "?"
            + urllib.parse.urlencode({"SAMLart": self.artifact})
        )

        with self.assertLogs("digid_eherkenning.backends", level="INFO") as log_watcher:
            response = self.client.get(url, follow=True)

        logs = [r.getMessage() for r in log_watcher.records]
        self.assertIn(
            "A technical error occurred from 127.0.0.1 during DigiD login.", logs
        )

        self.assertEqual(response.redirect_chain, [("/admin/login/", 302)])
        self.assertEqual(
            list(response.context["messages"])[0].message,
            _(
                "An error occurred in the communication with DigiD. "
                "Please try again later. If this error persists, please "
                "check the website https://www.digid.nl for the latest information."
            ),
        )

        # Make sure no user is created.
        self.assertEqual(User.objects.count(), 0)

    @skip("See issue #2. Not implemented")
    @responses.activate
    def test_invalid_subject_ip_address(self):
        root_element = etree.fromstring(self.artifact_response_soap)
        status_code = get_saml_element(
            root_element, "//saml:AuthnStatement/saml:SubjectLocality"
        )
        # We do the request with 127.0.0.1
        status_code.set("Address", "127.0.0.2")

        responses.add(
            responses.POST,
            "https://was-preprod1.digid.nl/saml/idp/resolve_artifact",
            body=etree.tostring(root_element),
            status=200,
        )

        url = (
            reverse("digid:acs")
            + "?"
            + urllib.parse.urlencode({"SAMLart": self.artifact})
        )

        with self.assertLogs("digid_eherkenning.backends", level="INFO") as log_watcher:
            response = self.client.get(url, follow=True)

        logs = [r.getMessage() for r in log_watcher.records]
        self.assertIn(
            "A technical error occurred from 127.0.0.1 during DigiD login.", logs
        )

        self.assertEqual(response.redirect_chain, [("/admin/login/", 302)])
        self.assertEqual(
            list(response.context["messages"])[0].message,
            "Login to DigiD did not succeed. Please try again.",
        )

        # Make sure no user is created.
        self.assertEqual(User.objects.count(), 0)

    @responses.activate
    def test_get(self):
        responses.add(
            responses.POST,
            "https://was-preprod1.digid.nl/saml/idp/resolve_artifact",
            body=self.artifact_response_soap,
            status=200,
        )

        url = (
            reverse("digid:acs")
            + "?"
            + urllib.parse.urlencode({"SAMLart": self.artifact, "RelayState": "/home/"})
        )
        with self.assertLogs("digid_eherkenning.backends", level="INFO") as log_watcher:
            response = self.client.get(url, secure=True)

        logs = [r.getMessage() for r in log_watcher.records]
        self.assertIn(
            "User user-12345678 (new account) from 127.0.0.1 logged in using DigiD",
            logs,
        )

        # Make sure we're redirect the the right place.
        self.assertEqual(response.url, "/home/")

        # Make sure the user is created and logged in.
        user = auth.get_user(self.client)
        self.assertEqual(user.username, "user-12345678")
        self.assertEqual(user.bsn, "12345678")

        # DigiD - Stap 6

        # 1.5 Voorbeeldbericht bij Stap 6 : Artifact Resolve (SOAP)
        # <samlp:ArtifactResolve
        #  xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol"
        #  xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion"
        #  xmlns:ds="http://www.w3.org/2000/09/xmldsig#"
        #  xmlns:ec="http://www.w3.org/2001/10/xml-exc-c14n#"
        #  ID="_1330416073" Version="2.0" IssueInstant="2012-02-28T09:01:13Z">
        #  <saml:Issuer>http://sp.example.com</saml:Issuer>
        #  <ds:Signature><!—Zie XML Signature--></ds:Signature>
        #  <samlp:Artifact>AAQAAMh48/1oXIMRdUmllwn9jJHyEgIi8=</samlp:Artifact>
        # </samlp:ArtifactResolve>

        tree = etree.fromstring(responses.calls[0].request.body)
        elements = tree.xpath(
            "//xmldsig:SignatureValue",
            namespaces={"xmldsig": "http://www.w3.org/2000/09/xmldsig#"},
        )
        elements[0].text = ""

        elements = tree.xpath(
            "//xmldsig:DigestValue",
            namespaces={"xmldsig": "http://www.w3.org/2000/09/xmldsig#"},
        )
        elements[0].text = ""

        elements = tree.xpath(
            "//xmldsig:X509Certificate",
            namespaces={"xmldsig": "http://www.w3.org/2000/09/xmldsig#"},
        )
        elements[0].text = ""

        elements = tree.xpath(
            "//samlp:Artifact",
            namespaces={"samlp": "urn:oasis:names:tc:SAML:2.0:protocol"},
        )

        # Make sure the Artifact is sent as-is.
        self.assertEqual(elements[0].text, self.artifact.decode("utf-8"))

        elements = tree.xpath(
            "//saml:Issuer",
            namespaces={"saml": "urn:oasis:names:tc:SAML:2.0:assertion"},
        )
        self.assertEqual(elements[0].text, "sp.example.nl/digid")

        # Make sure that the cache is checked for the InResponseTo returned
        # by the IDP.
        self.cache_mock.get.assert_called_once_with("digid__7afa5ce49")

    @override_settings(LOGIN_URL="/dummy/login")
    @responses.activate
    def test_no_authn_request(self):
        """
        Make sure that when the InResponseTo in the Response does not match
        any id we've given out, an error occurs.
        """
        self.cache_mock.get.return_value = None
        responses.add(
            responses.POST,
            "https://was-preprod1.digid.nl/saml/idp/resolve_artifact",
            body=self.artifact_response_soap,
            status=200,
        )

        url = (
            reverse("digid:acs")
            + "?"
            + urllib.parse.urlencode({"SAMLart": self.artifact})
        )
        response = self.client.get(url)

        with self.assertLogs("digid_eherkenning.backends", level="INFO") as log_watcher:
            response = self.client.get(url, secure=True)

        logs = [r.getMessage() for r in log_watcher.records]
        self.assertIn(
            "A technical error occurred from 127.0.0.1 during DigiD login.", logs
        )

        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/dummy/login")

        # Make sure no user is created.
        self.assertEqual(User.objects.count(), 0)

    @responses.activate
    def test_redirect_default(self):
        """
        Make sure the view returns to the default URL if no RelayState is set
        """
        responses.add(
            responses.POST,
            "https://was-preprod1.digid.nl/saml/idp/resolve_artifact",
            body=self.artifact_response_soap,
            status=200,
        )

        url = (
            reverse("digid:acs")
            + "?"
            + urllib.parse.urlencode({"SAMLart": self.artifact})
        )
        response = self.client.get(url)

        self.assertEqual(response.url, settings.LOGIN_REDIRECT_URL)

    @responses.activate
    def test_lower_session_age(self):
        """
        Make sure the session age is lowered. Since 'session_age' is
        set to 15 * 60 minutes in the configuration.

        DigiD requires a session of max 15 minutes. See DigiDCheck 2.2 T14 -- Sessieduur
        """
        responses.add(
            responses.POST,
            "https://was-preprod1.digid.nl/saml/idp/resolve_artifact",
            body=self.artifact_response_soap,
            status=200,
        )

        url = (
            reverse("digid:acs")
            + "?"
            + urllib.parse.urlencode({"SAMLart": self.artifact})
        )
        response = self.client.get(url)

        self.assertEqual(self.client.session.get_expiry_age(), 900)

    @responses.activate
    def test_user_cancels(self):
        """
        Test that when a user cancels this is logged properly.
        """
        root_element = etree.fromstring(self.artifact_response_soap)

        # Remove Assertion element. It will not be returned
        # when user cancels.
        assertion = get_saml_element(
            root_element,
            "//saml:Assertion",
        )
        assertion.getparent().remove(assertion)

        status_code = get_saml_element(
            root_element, "//samlp:Response/samlp:Status/samlp:StatusCode"
        )
        status_code.set("Value", "urn:oasis:names:tc:SAML:2.0:status:Responder")

        status_code.insert(
            0,
            etree.Element(
                "{urn:oasis:names:tc:SAML:2.0:protocol}StatusCode",
                Value="urn:oasis:names:tc:SAML:2.0:status:AuthnFailed",
            ),
        )

        responses.add(
            responses.POST,
            "https://was-preprod1.digid.nl/saml/idp/resolve_artifact",
            body=etree.tostring(root_element),
            status=200,
        )

        url = (
            reverse("digid:acs")
            + "?"
            + urllib.parse.urlencode({"SAMLart": self.artifact})
        )
        with self.assertLogs("digid_eherkenning.backends", level="INFO") as log_watcher:
            response = self.client.get(url, follow=True)

        logs = [r.getMessage() for r in log_watcher.records]
        self.assertIn(
            "The DigiD login from 127.0.0.1 did not succeed or was cancelled.", logs
        )

        self.assertEqual(response.redirect_chain, [("/admin/login/", 302)])
        self.assertEqual(
            list(response.context["messages"])[0].message,
            _("You have cancelled logging in with DigiD."),
        )

        # Make sure no user is created.
        self.assertEqual(User.objects.count(), 0)



from django.core.validators import RegexValidator
from django.utils.translation import gettext_lazy as _

# See `OINType` in eherkenning-dc.xml XSD
oin_validator = RegexValidator(
    regex=r"[0-9]{20}",
    message=_("A valid OIN consists of 20 digits."),
)



from .digid import (  # noqa
    DigiDAssertionConsumerServiceView,
    DigiDLoginView,
    DigiDLogoutView,
    DigidSingleLogoutRedirectView,
    DigidSingleLogoutSoapView,
)
from .eherkenning import (  # noqa
    eHerkenningAssertionConsumerServiceView,
    eHerkenningLoginView,
)



from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.utils.translation import gettext_lazy as _

from ..settings import get_setting
from .base import BaseConfiguration


def default_digid_requested_attributes():
    return [
        {
            "name": "bsn",
            "required": True,
        }
    ]


class DigidConfiguration(BaseConfiguration):
    attribute_consuming_service_index = models.CharField(
        _("Attribute consuming service index"),
        blank=True,
        default="1",
        help_text=_("Attribute consuming service index"),
        max_length=100,
    )
    requested_attributes = models.JSONField(
        _("requested attributes"),
        default=default_digid_requested_attributes,
        help_text=_(
            "A list of strings (or objects) with the requested attributes, e.g. '[\"bsn\"]'"
        ),
    )
    slo = models.BooleanField(
        _("Single logout"),
        default=True,
        help_text=_("If enabled, Single Logout is supported"),
    )

    class Meta:
        verbose_name = _("Digid configuration")

    def as_dict(self) -> dict:
        """
        Emit the configuration as a dictionary compatible with the old settings format.
        """
        organization = None

        if (
            not self.certificate
            or not self.certificate.private_key
            or not self.certificate.public_certificate
        ):
            raise ImproperlyConfigured(
                "No (valid) certificate configured. The configuration needs a "
                "certificate with private key and public certificate."
            )

        if self.organization_url and self.organization_name:
            organization = {
                "nl": {
                    "name": self.organization_name,
                    "displayname": self.organization_name,
                    "url": self.organization_url,
                }
            }

        return {
            "base_url": self.base_url,
            "entity_id": self.entity_id,
            "metadata_file": self.idp_metadata_file,
            "key_file": self.certificate.private_key,
            "cert_file": self.certificate.public_certificate,
            "service_entity_id": self.idp_service_entity_id,
            "attribute_consuming_service_index": self.attribute_consuming_service_index,
            "service_name": self.service_name,
            "service_description": self.service_description,
            "requested_attributes": self.requested_attributes or [],
            # optional in runtime code
            "want_assertions_encrypted": self.want_assertions_encrypted,
            "want_assertions_signed": self.want_assertions_signed,
            "key_passphrase": self.key_passphrase or None,
            "signature_algorithm": self.signature_algorithm,
            "digest_algorithm": self.digest_algorithm or None,
            "technical_contact_person_telephone": self.technical_contact_person_telephone
            or None,
            "technical_contact_person_email": self.technical_contact_person_email
            or None,
            "organization": organization,
            "session_age": get_setting("DIGID_SESSION_AGE"),
            "slo": self.slo,
            "artifact_resolve_content_type": self.artifact_resolve_content_type,
        }



from django.test import TestCase
from django.urls import reverse

import pytest

from digid_eherkenning.models import DigidConfiguration, EherkenningConfiguration
from digid_eherkenning.saml2.digid import DigiDClient
from digid_eherkenning.saml2.eherkenning import eHerkenningClient


@pytest.mark.usefixtures("digid_config_defaults", "temp_private_root")
class DigidClientTests(TestCase):
    def test_wants_assertions_signed_setting_default(self):
        config = DigidConfiguration.get_solo()

        conf = config.as_dict()
        conf.setdefault("acs_path", reverse("digid:acs"))

        digid_client = DigiDClient()
        config_dict = digid_client.create_config_dict(conf)

        self.assertIn("wantAssertionsSigned", config_dict["security"])
        self.assertTrue(config_dict["security"]["wantAssertionsSigned"])

    def test_wants_assertions_signed_setting_changed(self):
        config = DigidConfiguration.get_solo()
        config.want_assertions_signed = False
        config.save()

        conf = config.as_dict()
        conf.setdefault("acs_path", reverse("digid:acs"))

        digid_client = DigiDClient()
        config_dict = digid_client.create_config_dict(conf)

        self.assertIn("wantAssertionsSigned", config_dict["security"])
        self.assertFalse(config_dict["security"]["wantAssertionsSigned"])

    def test_artifact_resolve_content_type_settings_default(self):
        config = DigidConfiguration.get_solo()

        conf = config.as_dict()
        conf.setdefault("acs_path", reverse("digid:acs"))

        digid_client = DigiDClient()
        config_dict = digid_client.create_config_dict(conf)

        self.assertIn("resolveArtifactBindingContentType", config_dict["idp"])
        self.assertIn(
            "application/soap+xml",
            config_dict["idp"]["resolveArtifactBindingContentType"],
        )

    def test_artifact_resolve_content_type_settings(self):
        config = DigidConfiguration.get_solo()
        config.artifact_resolve_content_type = "text/xml"
        config.save()

        conf = config.as_dict()
        conf.setdefault("acs_path", reverse("digid:acs"))

        digid_client = DigiDClient()
        config_dict = digid_client.create_config_dict(conf)

        self.assertIn("resolveArtifactBindingContentType", config_dict["idp"])
        self.assertIn(
            "text/xml", config_dict["idp"]["resolveArtifactBindingContentType"]
        )


@pytest.mark.usefixtures("eherkenning_config_defaults", "temp_private_root")
class EHerkenningClientTests(TestCase):
    def test_wants_assertions_signed_setting_default(self):
        config = EherkenningConfiguration.get_solo()

        conf = config.as_dict()
        conf.setdefault("acs_path", reverse("eherkenning:acs"))

        eherkenning_client = eHerkenningClient()
        config_dict = eherkenning_client.create_config_dict(conf)

        self.assertIn("wantAssertionsSigned", config_dict["security"])
        self.assertTrue(config_dict["security"]["wantAssertionsSigned"])

    def test_wants_assertions_signed_setting_changed(self):
        config = EherkenningConfiguration.get_solo()
        config.want_assertions_signed = False
        config.save()

        conf = config.as_dict()
        conf.setdefault("acs_path", reverse("eherkenning:acs"))

        eherkenning_client = eHerkenningClient()
        config_dict = eherkenning_client.create_config_dict(conf)

        self.assertIn("wantAssertionsSigned", config_dict["security"])
        self.assertFalse(config_dict["security"]["wantAssertionsSigned"])

    def test_signature_digest_algorithm_settings_changed(self):
        config = EherkenningConfiguration.get_solo()
        config.signature_algorithm = "http://www.w3.org/2001/04/xmldsig-more#rsa-sha256"
        config.digest_algorithm = "http://www.w3.org/2001/04/xmlenc#sha256"
        config.save()

        conf = config.as_dict()
        conf.setdefault("acs_path", reverse("eherkenning:acs"))

        eherkenning_client = eHerkenningClient()
        config_dict = eherkenning_client.create_config_dict(conf)

        self.assertIn("signatureAlgorithm", config_dict["security"])
        self.assertIn("digestAlgorithm", config_dict["security"])
        self.assertEqual(
            "http://www.w3.org/2001/04/xmldsig-more#rsa-sha256",
            config_dict["security"]["signatureAlgorithm"],
        )
        self.assertEqual(
            "http://www.w3.org/2001/04/xmlenc#sha256",
            config_dict["security"]["digestAlgorithm"],
        )

    def test_artifact_resolve_content_type_settings_default(self):
        config = EherkenningConfiguration.get_solo()

        conf = config.as_dict()
        conf.setdefault("acs_path", reverse("eherkenning:acs"))

        eherkenning_client = eHerkenningClient()
        config_dict = eherkenning_client.create_config_dict(conf)

        self.assertIn("resolveArtifactBindingContentType", config_dict["idp"])
        self.assertIn(
            "application/soap+xml",
            config_dict["idp"]["resolveArtifactBindingContentType"],
        )

    def test_artifact_resolve_content_type_settings(self):
        config = EherkenningConfiguration.get_solo()
        config.artifact_resolve_content_type = "text/xml"
        config.save()

        conf = config.as_dict()
        conf.setdefault("acs_path", reverse("eherkenning:acs"))

        eherkenning_client = eHerkenningClient()
        config_dict = eherkenning_client.create_config_dict(conf)

        self.assertIn("resolveArtifactBindingContentType", config_dict["idp"])
        self.assertIn(
            "text/xml", config_dict["idp"]["resolveArtifactBindingContentType"]
        )



from django.core.exceptions import ValidationError
from django.db import models
from django.utils.encoding import force_str
from django.utils.translation import gettext_lazy as _

from privates.fields import PrivateMediaFileField
from simple_certmanager.models import Certificate
from solo.models import SingletonModel

from ..choices import DigestAlgorithms, SignatureAlgorithms, XMLContentTypes


class ConfigurationManager(models.Manager):
    def get_queryset(self):
        qs = super().get_queryset()
        return qs.select_related("certificate")


class BaseConfiguration(SingletonModel):
    certificate = models.ForeignKey(
        Certificate,
        null=True,
        on_delete=models.PROTECT,
        verbose_name=_("key pair"),
        help_text=_(
            "The private key and public certificate pair to use during the "
            "authentication flow."
        ),
    )
    idp_metadata_file = PrivateMediaFileField(
        _("identity provider metadata"),
        blank=False,
        help_text=_("The metadata file of the identity provider."),
    )
    idp_service_entity_id = models.CharField(
        _("identity provider service entity ID"),
        max_length=255,
        blank=False,
        help_text=_(
            "Example value: 'https://was-preprod1.digid.nl/saml/idp/metadata'. Note "
            "that this must match the 'entityID' attribute on the "
            "'md:EntityDescriptor' node found in the Identity Provider's metadata."
        ),
    )
    want_assertions_signed = models.BooleanField(
        _("want assertions signed"),
        default=True,
        help_text=_(
            "If True, the XML assertions need to be signed, otherwise the whole "
            "response needs to be signed."
        ),
        max_length=100,
    )
    want_assertions_encrypted = models.BooleanField(
        _("want assertions encrypted"),
        default=False,
        help_text=_("If True the XML assertions need to be encrypted."),
        max_length=100,
    )
    artifact_resolve_content_type = models.CharField(
        _("resolve artifact binding content type"),
        choices=XMLContentTypes.choices,
        default=XMLContentTypes.soap_xml,
        max_length=100,
        help_text=_(
            "'application/soap+xml' is considered legacy and modern brokers typically "
            "expect 'text/xml'."
        ),
    )
    key_passphrase = models.CharField(
        _("key passphrase"),
        blank=True,
        help_text=_("Passphrase for the private key used by the SOAP client."),
        max_length=100,
    )
    signature_algorithm = models.CharField(
        _("signature algorithm"),
        blank=True,
        choices=SignatureAlgorithms.choices,
        default=SignatureAlgorithms.rsa_sha1,
        help_text=_(
            "Signature algorithm. Note that DSA_SHA1 and RSA_SHA1 are deprecated, but "
            "RSA_SHA1 is still the default value in the SAMLv2 standard. Warning: "
            "there are known issues with single-logout functionality if using anything "
            "other than SHA1 due to some hardcoded algorithm."
        ),
        max_length=100,
    )
    digest_algorithm = models.CharField(
        _("digest algorithm"),
        blank=True,
        choices=DigestAlgorithms.choices,
        default=DigestAlgorithms.sha1,
        help_text=_(
            "Digest algorithm. Note that SHA1 is deprecated, but still the default "
            "value in the SAMLv2 standard. Warning: "
            "there are known issues with single-logout functionality if using anything "
            "other than SHA1 due to some hardcoded algorithm."
        ),
        max_length=100,
    )
    entity_id = models.CharField(
        _("entity ID"), help_text=_("Service provider entity ID."), max_length=100
    )
    base_url = models.URLField(
        _("base URL"),
        help_text=_("Base URL of the application, without trailing slash."),
        max_length=100,
    )
    service_name = models.CharField(
        _("service name"),
        help_text=_("Name of the service you are providing."),
        max_length=100,
    )
    service_description = models.CharField(
        _("service description"),
        help_text=_("A description of the service you are providing."),
        max_length=100,
    )
    technical_contact_person_telephone = models.CharField(
        _("technical contact: phone number"),
        blank=True,
        help_text=_(
            "Telephone number of the technical person responsible for this "
            "DigiD/eHerkenning/eIDAS setup. For it to show up in the metata, you "
            "should also specify the email address."
        ),
        max_length=100,
    )
    technical_contact_person_email = models.CharField(
        _("technical contact: email"),
        blank=True,
        help_text=_(
            "Email address of the technical person responsible for this "
            "DigiD/eHerkenning/eIDAS setup. For it to show up in the metadata, you "
            "should also specify the phone number."
        ),
        max_length=100,
    )
    organization_url = models.URLField(
        _("organization URL"),
        blank=True,
        help_text=_(
            "URL of the organization providing the service for which "
            "DigiD/eHerkenning/eIDAS login is configured. For it to show up in the "
            "metadata, you should also specify the organization name."
        ),
        max_length=255,
    )
    organization_name = models.CharField(
        _("organization name"),
        blank=True,
        help_text=_(
            "URL of the organization providing the service for which "
            "DigiD/eHerkenning/eIDAS login is configured. For it to show up in the "
            "metadata, you should also specify the organization URL."
        ),
        max_length=100,
    )

    objects = ConfigurationManager()

    class Meta:
        abstract = True

    def __str__(self):
        return force_str(self._meta.verbose_name)

    def save(self, *args, **kwargs):
        if self.base_url.endswith("/"):
            self.base_url = self.base_url[:-1]
        super().save(*args, **kwargs)

    def clean(self):
        if not self.certificate:
            raise ValidationError(_("You must select a certificate"))
        super().clean()



from django.utils import timezone

from ...models import EherkenningConfiguration
from ...saml2.eherkenning import generate_dienst_catalogus_metadata
from .generate_eherkenning_metadata import Command as EherkenningCommand


def _remove_action_by_dest(parser, dest: str):
    for action in parser._actions:
        if action.dest != dest:
            continue
        parser._remove_action(action)
        break

    for action in parser._action_groups:
        for group_action in action._group_actions:
            if group_action.dest != dest:
                continue
            action._group_actions.remove(group_action)
            return


class Command(EherkenningCommand):
    help = "Create the eHerkenning dienstcatalogus file"

    def add_arguments(self, parser):
        super().add_arguments(parser)

        # delete arguments that we don't use
        dests_to_delete = [
            "want_assertions_encrypted",
            "want_assertions_signed",
            "key_passphrase",
            "technical_contact_person_telephone",
            "technical_contact_person_email",
            "organization_url",
        ]
        # remove actions not relevant for this command, but still re-use the bulk
        # from the eherkenning metadata generation command
        for dest in dests_to_delete:
            _remove_action_by_dest(parser, dest)

        config: EherkenningConfiguration = self._get_config()

        parser.add_argument(
            "--privacy-policy",
            required=not config.privacy_policy,
            help=(
                "The URL where the privacy policy from the organisation providing the "
                "service can be found."
            ),
        )
        parser.add_argument(
            "--makelaar-id",
            required=not config.makelaar_id,
            help="OIN of the broker used to set up eHerkenning/eIDAS.",
        )

    def get_filename(self):
        date_string = timezone.now().date().isoformat()
        return f"eherkenning-dienstcatalogus-{date_string}.xml"

    def generate_metadata(self, options):
        return generate_dienst_catalogus_metadata()



from digid_eherkenning.saml2.base import create_saml2_request


def test_saml2_get_request_creation(rf):
    django_request = rf.get("/irrelevant", {"some_query": "string"})
    base_url = "https://example.com:8443/foo"

    saml_request = create_saml2_request(base_url, django_request)

    assert saml_request == {
        "https": "on",
        "http_host": "example.com:8443",
        "script_name": "/irrelevant",
        "get_data": {"some_query": ["string"]},
        "post_data": {},
        "query_string": "some_query=string",
        "body": b"",
    }


def test_saml2_post_request_creation(rf):
    django_request = rf.post("/irrelevant", data={"some_query": "string"})
    django_request.body
    base_url = "http://example.com/foo"

    saml_request = create_saml2_request(base_url, django_request)

    assert saml_request == {
        "https": "off",
        "http_host": "example.com",
        "script_name": "/irrelevant",
        "get_data": {},
        "post_data": {"some_query": ["string"]},
        "query_string": "",
        "body": (
            b'--BoUnDaRyStRiNg\r\nContent-Disposition: form-data; name="some_query"'
            b"\r\n\r\nstring\r\n--BoUnDaRyStRiNg--\r\n"
        ),
    }



from pathlib import Path
from typing import Sequence, Type

from django.core.files import File
from django.core.management.base import BaseCommand
from django.db import transaction

from simple_certmanager.constants import CertificateTypes
from simple_certmanager.models import Certificate

from ...models.base import BaseConfiguration

try:
    from argparse import BooleanOptionalAction
except ImportError:
    from ..utils import BooleanOptionalAction


class SamlMetadataBaseCommand(BaseCommand):
    config_model: Type[BaseConfiguration]
    default_certificate_label: str

    def add_arguments(self, parser):
        """
        Add arguments that map to configuration model fields.

        You can use a different flag, but then ensure the ``dest`` kwarg is specified.
        Options are applied to the specified configuration model instance if a model
        field with the same name as the option exists.
        """
        # check current config to determine if an option is required or not
        config = self._get_config()
        has_private_key = config.certificate and config.certificate.private_key
        has_certificate = config.certificate and config.certificate.public_certificate

        parser.add_argument(
            "--want-assertions-encrypted",
            action="store_true",
            help="If True the XML assertions need to be encrypted. Defaults to False",
        )
        parser.add_argument(
            "--no-only-assertions-signed",
            dest="want_assertions_signed",
            action="store_false",
            help=(
                "If True, the XML assertions need to be signed, otherwise the whole "
                "response needs to be signed. Defaults to only assertions signed."
            ),
        )
        parser.add_argument(
            "--key-file",
            required=not has_private_key,
            help=(
                "The filepath to the TLS key. This will be used both by the SOAP "
                "client and for signing the requests."
            ),
        )
        parser.add_argument(
            "--cert-file",
            required=not has_certificate,
            help=(
                "The filepath to the TLS certificate. This will be used both by the "
                "SOAP client and for signing the requests."
            ),
        )
        parser.add_argument(
            "--key-passphrase",
            help="Passphrase for SOAP client",
            default=None,
        )
        parser.add_argument(
            "--signature-algorithm",
            help="Signature algorithm, defaults to RSA_SHA1",
            default="http://www.w3.org/2000/09/xmldsig#rsa-sha1",
        )
        parser.add_argument(
            "--digest-algorithm",
            help="Digest algorithm, defaults to SHA1",
            default="http://www.w3.org/2000/09/xmldsig#sha1",
        )
        parser.add_argument(
            "--entity-id",
            required=not config.entity_id,
            help="Service provider entity ID",
        )
        parser.add_argument(
            "--base-url",
            required=not config.base_url,
            help="Base URL of the application",
        )
        parser.add_argument(
            "--service-name",
            required=not config.service_name,
            help="The name of the service for which DigiD login is required",
        )
        parser.add_argument(
            "--service-description",
            required=not config.service_description,
            help="A description of the service for which DigiD login is required",
        )
        parser.add_argument(
            "--technical-contact-person-telephone",
            help=(
                "Telephone number of the technical person responsible for this DigiD "
                "setup. For it to be used, --technical-contact-person-email should "
                "also be set."
            ),
        )
        parser.add_argument(
            "--technical-contact-person-email",
            help=(
                "Email address of the technical person responsible for this DigiD "
                "setup. For it to be used, --technical-contact-person-telephone "
                "should also be set."
            ),
        )
        parser.add_argument(
            "--organization-name",
            help=(
                "Name of the organisation providing the service for which DigiD login "
                "is setup. For it to be used, also --organization-url should be filled."
            ),
        )
        parser.add_argument(
            "--organization-url",
            help=(
                "URL of the organisation providing the service for which DigiD login "
                "is setup. For it to be used, also --organization-name should be "
                "filled."
            ),
        )
        parser.add_argument(
            "--output-file",
            help=(
                "Name of the file to which to write the metadata. Otherwise will be "
                "printed on stdout"
            ),
        )
        parser.add_argument(
            "--test",
            "--debug",
            action="store_true",
            help="If True the metadata is printed to stdout. Defaults to False",
        )
        parser.add_argument(
            "--save-config",
            action=BooleanOptionalAction,
            required=True,
            help="Save the configuration overrides specified via the command line.",
        )

    def get_filename(self) -> str:  # pragma:nocover
        raise NotImplementedError

    def generate_metadata(self, options: dict) -> bytes:  # pragma:nocover
        raise NotImplementedError

    def _get_config(self):
        if not hasattr(self, "_config"):
            self._config = self.config_model.get_solo()
        return self._config

    def _set_certificate(self, config: BaseConfiguration, options: dict):
        certificate = config.certificate

        # no certificate exists yet -> create one
        if certificate is None:
            certificate = Certificate.objects.create(
                label=self.default_certificate_label,
                type=CertificateTypes.key_pair,
            )
            config.certificate = certificate

        # enforce that the specified key/certificate are used
        for option, filefield in (
            ("key_file", "private_key"),
            ("cert_file", "public_certificate"),
        ):
            filepath = options[option]
            if not filepath:
                continue

            path = Path(filepath)
            with path.open("rb") as infile:
                field_file = getattr(certificate, filefield)
                field_file.save(path.name, File(infile), save=False)

        certificate.save()

    @transaction.atomic
    def _generate_metadata(self, options: dict) -> bytes:
        valid_field_names = [f.name for f in self.config_model._meta.get_fields()]
        config = self._get_config()

        self._set_certificate(config, options)

        for key, value in options.items():
            if key not in valid_field_names:
                continue
            # optional, unspecified -> go with the model default or current value
            if value is None:
                continue
            setattr(config, key, value)

        config.save()

        metadata = self.generate_metadata(options)

        transaction.set_rollback(not options["save_config"])

        return metadata

    def handle(self, *args, **options):
        metadata_content = self._generate_metadata(options)

        if options["test"]:
            self.stdout.write(metadata_content.decode("utf-8"))
            return

        filename = options["output_file"] or self.get_filename()
        with open(filename, "xb") as metadata_file:
            metadata_file.write(metadata_content)

        self.stdout.write(
            self.style.SUCCESS("Metadata file successfully generated: %s" % filename)
        )



import logging
from typing import Callable, Type

from django.core.exceptions import ValidationError
from django.http import HttpRequest, HttpResponse, HttpResponseBadRequest
from django.http.response import HttpResponseBase
from django.utils.translation import gettext as _
from django.views import View

from ..models.base import BaseConfiguration

logger = logging.getLogger(__name__)


class MetadataView(View):
    config_model: Type[BaseConfiguration] = BaseConfiguration
    metadata_generator: Callable[[], bytes] = lambda: b""

    def get(self, request: HttpRequest) -> HttpResponseBase:
        config = self.config_model.get_solo()

        try:
            config.clean()
        except ValidationError as error:
            logger.warning(
                "Invalid '%s' configuration",
                self.config_model._meta.verbose_name,
                exc_info=error,
            )
            return self._get_generic_error_response()

        try:
            metadata = self.metadata_generator()
        except Exception as error:
            logger.warning(
                "Failed generating metadata for '%s' at '%s'",
                self.config_model._meta.verbose_name,
                request.path,
                exc_info=error,
                extra={
                    "request": request,
                    "config_model": self.config_model,
                },
            )
            return self._get_generic_error_response()
        return HttpResponse(metadata, content_type="text/xml")

    @staticmethod
    def _get_generic_error_response() -> HttpResponseBadRequest:
        error_message = _(
            "Something went wrong while generating the metadata. Please get in touch "
            "with your technical contact person and inform them the configuration is "
            "invalid."
        )
        return HttpResponseBadRequest(error_message, content_type="text/plain")



from importlib import import_module

from django.conf import settings

from defusedxml.lxml import parse
from lxml import etree


def get_client_ip(request):
    x_forwarded_for = request.META.get("HTTP_X_FORWARDED_FOR")
    if x_forwarded_for:
        # The general format of the field is:  X-Forwarded-For: client, proxy1, proxy2
        ip = x_forwarded_for.split(",")[0]
    else:
        ip = request.META.get("REMOTE_ADDR")
    return ip


def validate_xml(xml, xsd):
    """
    Validates `xml` against the schema `xsd`

    :param xml: File-like object with XML to be validated
    :param xsd: File-like object containing the schema xml
    :return the error log if validation errors occurred
    """
    try:
        xmlschema = etree.XMLSchema(parse(xsd))
        doc = parse(xml)
        if not xmlschema.validate(doc):
            return xmlschema.error_log
        return None
    finally:
        if hasattr(xml, "seek"):
            xml.seek(0)


def logout_user(user):
    """
    forcefully logout user from their sessions
    """
    from sessionprofile.models import SessionProfile

    session_profiles = SessionProfile.objects.filter(user=user)
    SessionStore = import_module(settings.SESSION_ENGINE).SessionStore
    s = SessionStore()
    for sp in session_profiles:
        s.delete(sp.session_key)

    session_profiles.delete()



from unittest.mock import patch

from django.conf import settings
from django.test import TestCase
from django.urls import reverse, reverse_lazy

import pytest
from freezegun import freeze_time
from furl import furl
from lxml import etree
from onelogin.saml2.constants import OneLogin_Saml2_Constants
from onelogin.saml2.utils import OneLogin_Saml2_Utils
from onelogin.saml2.xml_utils import OneLogin_Saml2_XML
from sessionprofile.models import SessionProfile

from digid_eherkenning.choices import SectorType
from digid_eherkenning.models import DigidConfiguration

from .project.choices import UserLoginType
from .project.models import User
from .utils import get_saml_element


@pytest.mark.usefixtures("digid_config", "temp_private_root")
class DigidLogoutViewTests(TestCase):
    maxDiff = None

    @freeze_time("2020-04-09T08:31:46Z")
    @patch("onelogin.saml2.utils.uuid4")
    def test_logout(self, uuid_mock):
        """
        DigID

        Make sure DigiD - 4.2.2 Stap U2

        works as intended.
        """
        config = DigidConfiguration.get_solo()
        uuid_mock.hex = "80dd245883b84bd98dacbf3978af3d03"
        user = User.objects.create_user(
            username="testuser", password="test", bsn="12345670"
        )
        self.client.force_login(user)

        response = self.client.get(reverse("digid:logout"))

        self.assertEqual(response.status_code, 302)
        # user is still logged in
        self.assertEqual(self.client.session["_auth_user_id"], str(user.id))
        self.assertEqual(
            self.client.session["logout_request_id"],
            "ONELOGIN_5ba93c9db0cff93f52b521d7420e43f6eda2784f",
        )

        logout_url = response.url
        f = furl(logout_url)

        self.assertEqual(f.origin, "https://preprod1.digid.nl")
        self.assertEqual(f.path, "/saml/idp/request_logout")
        self.assertEqual(
            f.args.keys(), ["SAMLRequest", "RelayState", "Signature", "SigAlg"]
        )
        self.assertEqual(f.args["RelayState"], settings.LOGOUT_REDIRECT_URL)

        # check SAML request
        saml_request = OneLogin_Saml2_Utils.decode_base64_and_inflate(
            f.args["SAMLRequest"]
        )

        # Digid 2.1 Voorbeeldbericht bij Stap U2: Logout Request
        # Dit is een http redirect bericht. De signing wordt in de URI meegezonden.
        #
        # <?xml version="1.0" encoding="UTF-8"?>
        # <samlp:LogoutRequest
        #   xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol"
        #   xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion"
        #   ID="_1330416516" Version="2.0" IssueInstant="2012-02-28T09:08:36Z">
        #   <saml:Issuer>http://sp.example.com</saml:Issuer>
        #   <saml:NameID>s00000000:12345678</saml:NameID>
        # </samlp:LogoutRequest>

        expected_request = (
            "<samlp:LogoutRequest "
            'xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol" '
            'xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion" '
            'ID="ONELOGIN_5ba93c9db0cff93f52b521d7420e43f6eda2784f" Version="2.0" IssueInstant="2020-04-09T08:31:46Z" '
            'Destination="https://preprod1.digid.nl/saml/idp/request_logout">'
            f"<saml:Issuer>{config.entity_id}</saml:Issuer>"
            f"<saml:NameID>{SectorType.bsn}:{user.bsn}</saml:NameID>"
            "</samlp:LogoutRequest>"
        )
        self.assertXMLEqual(
            etree.tostring(etree.fromstring(saml_request), pretty_print=True).decode(
                "utf-8"
            ),
            etree.tostring(
                etree.fromstring(expected_request), pretty_print=True
            ).decode("utf-8"),
        )

        # check signature algorithm
        self.assertEqual(f.args["SigAlg"], OneLogin_Saml2_Constants.RSA_SHA1)
        self.assertIsNotNone(f.args["Signature"])

    def test_logout_without_bsn(self):
        """
        check that only users with nameId can logout
        """
        user = User.objects.create_user(username="testuser", password="test", bsn="")
        self.client.force_login(user)

        response = self.client.get(reverse("digid:logout"))

        self.assertEqual(response.status_code, 403)


@pytest.mark.usefixtures("digid_config", "temp_private_root")
class DigidSloLogoutRedirectTests(TestCase):
    url = reverse_lazy("digid:slo-redirect")

    def setUp(self):
        super().setUp()

        self.user = User.objects.create_user(
            username="testuser", password="test", bsn="12345670"
        )
        self.client.force_login(self.user)

        # 2.4 Voorbeeldbericht bij Stap U5: Response Redirect
        # Dit is een http redirect bericht. De signing wordt in de URI meegezonden.
        #
        # <?xml version="1.0"?>
        # <samlp:LogoutResponse
        #   xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion"
        #   xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol"
        #   xmlns:ds="http://www.w3.org/2000/09/xmldsig#"
        #   xmlns:ec="http://www.w3.org/2001/10/xml-exc-c14n#"
        #   Version="2.0" Destination="" InResponseTo="_43faa9487043be"
        #   ID="_882ff30b891047ca111" IssueInstant="2011-08-31T08:57:47Z">
        #   <saml:Issuer>https://idp.example.com</saml:Issuer>
        #   <samlp:Status>
        #     <samlp:StatusCode Value="urn:oasis:names:tc:SAML:2.0:status:Success"/>
        #   </samlp:Status>
        # </samlp:LogoutResponse>

        self.logout_response = (
            '<?xml version="1.0"?>'
            "<samlp:LogoutResponse "
            'xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion" '
            'xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol" '
            'xmlns:ds="http://www.w3.org/2000/09/xmldsig#" '
            'xmlns:ec="http://www.w3.org/2001/10/xml-exc-c14n#" '
            'Version="2.0" Destination="" InResponseTo="_43faa9487043be" '
            'ID="_882ff30b891047ca111" IssueInstant="2011-08-31T08:57:47Z">'
            "<saml:Issuer>https://was-preprod1.digid.nl/saml/idp/metadata</saml:Issuer>"
            "<samlp:Status>"
            '<samlp:StatusCode Value="urn:oasis:names:tc:SAML:2.0:status:Success"/>'
            "</samlp:Status>"
            "</samlp:LogoutResponse>"
        )

        self.validate_sign_patcher = patch.object(
            OneLogin_Saml2_Utils, "validate_binary_sign"
        )
        self.validate_sign_mock = self.validate_sign_patcher.start()

        self.addCleanup(patch.stopall)

    def test_logout_response(self):
        logout_response_encoded = OneLogin_Saml2_Utils.deflate_and_base64_encode(
            self.logout_response
        )
        data = {
            "SAMLResponse": logout_response_encoded,
            "RelayState": "/",
            "SigAlg": "http://www.w3.org/2000/09/xmldsig#rsa-sha1&",
            "Signature": "",
        }

        with self.assertLogs("digid_eherkenning.views", level="INFO") as log_watcher:
            response = self.client.get(self.url, data)

        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/")
        # local logout is done
        self.assertFalse("_auth_user_id" in self.client.session)

        logs = [r.getMessage() for r in log_watcher.records]
        self.assertIn(f"User {self.user} has successfully logged out of Digid", logs)

    def test_logout_response_status_code_failed(self):
        #  modify status code
        root_element = etree.fromstring(self.logout_response)
        status_code = get_saml_element(
            root_element, "//samlp:LogoutResponse/samlp:Status/samlp:StatusCode"
        )
        status_code.set("Value", "urn:oasis:names:tc:SAML:2.0:status:Responder")
        status_code.insert(
            0,
            etree.Element(
                "{urn:oasis:names:tc:SAML:2.0:protocol}StatusCode",
                Value="urn:oasis:names:tc:SAML:2.0:status:RequestDenied",
            ),
        )
        logout_response_encoded = OneLogin_Saml2_Utils.deflate_and_base64_encode(
            etree.tostring(root_element)
        )
        data = {
            "SAMLResponse": logout_response_encoded,
            "RelayState": "/",
            "SigAlg": "http://www.w3.org/2000/09/xmldsig#rsa-sha1&",
            "Signature": "",
        }

        with self.assertLogs("digid_eherkenning.views", level="INFO") as log_watcher:
            response = self.client.get(self.url, data)

        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/")
        # local logout is not done
        self.assertTrue("_auth_user_id" in self.client.session)

        logs = [r.getMessage() for r in log_watcher.records]
        self.assertIn(
            "An error occurred during logout from Digid: logout_not_success",
            logs,
        )


@pytest.mark.usefixtures("digid_config", "temp_private_root")
class DigidSloLogoutSOAPRequestTests(TestCase):
    maxDiff = None
    url = reverse_lazy("digid:slo-soap")

    def setUp(self):
        super().setUp()

        self.user = User.objects.create_user(
            username="testuser",
            password="test",
            bsn="12345670",
            login_type=UserLoginType.digid,
        )
        self.client.force_login(self.user)
        SessionProfile.objects.create(
            user=self.user, session_key=self.client.session.session_key
        )

        # 2.2 Voorbeeldbericht bij Stap U3: LogoutRequest (SOAP)
        # In een Soap envelope.
        # <samlp:LogoutRequest
        #   xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol"
        #   xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion"
        #   xmlns:ds="http://www.w3.org/2000/09/xmldsig#"
        #   xmlns:ec="http://www.w3.org/2001/10/xml-exc-c14n#"
        #   ID="_1331125262" Version="2.0" IssueInstant="2012-03-07T14:01:02Z">
        #   <saml:Issuer> http://sp.example.com</saml:Issuer>.
        #   <ds:Signature> <!-- See XML Signature --> </ds:Signature>.
        #   <saml:NameID> s00000000:12345678</saml:NameID>
        # </samlp:LogoutRequest>.
        #
        # 1.2 Soap Envelope
        # Al het Saml back-channel verkeer wordt in een Soap envelope geplaatst
        # <?xml version="1.0" encoding="UTF-8"?>
        # <soapenv:Envelope
        #   xmlns:soapenv=http://schemas.xmlsoap.org/soap/envelope/
        #   xmlns:xsd="http://www.w3.org/2001/XMLSchema"
        #   xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
        #   <soapenv:Body>
        #     <!—SAML BERICHT -->
        #   </soapenv:Body>
        # </soapenv:Envelope>

        self.signature = (
            "<ds:Signature>"
            "<ds:SignedInfo>"
            '<ds:CanonicalizationMethod Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#"/>'
            '<ds:SignatureMethod Algorithm="http://www.w3.org/2001/04/xmldsig-more#rsa-sha256"/>'
            '<ds:Reference URI="#_1331125262">'
            "<ds:Transforms>"
            '<ds:Transform Algorithm="http://www.w3.org/2000/09/xmldsig#enveloped-signature"/>'
            '<ds:Transform Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#">'
            '<ec:InclusiveNamespaces PrefixList="ds saml samlp xs"/>'
            "</ds:Transform>"
            "</ds:Transforms>"
            '<ds:DigestMethod Algorithm="http://www.w3.org/2001/04/xmlenc#sha256"/>'
            "<ds:DigestValue></ds:DigestValue>"
            "</ds:Reference>"
            "</ds:SignedInfo>"
            "<ds:SignatureValue></ds:SignatureValue>"
            "</ds:Signature>"
        )

        self.logout_request = (
            "<samlp:LogoutRequest "
            'xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol" '
            'xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion" '
            'xmlns:ds="http://www.w3.org/2000/09/xmldsig#" '
            'xmlns:ec="http://www.w3.org/2001/10/xml-exc-c14n#" '
            'ID="_1331125262" Version="2.0" IssueInstant="2012-03-07T14:01:02Z">'
            "<saml:Issuer>https://was-preprod1.digid.nl/saml/idp/metadata</saml:Issuer>"
            + self.signature
            + "<saml:NameID>s00000000:12345670</saml:NameID>"
            "</samlp:LogoutRequest>"
        )
        self.logout_request_soap = (
            "<soapenv:Envelope "
            'xmlns:soapenv="http://schemas.xmlsoap.org/soap/envelope/" '
            'xmlns:xsd="http://www.w3.org/2001/XMLSchema" '
            'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'
            "<soapenv:Body>" + self.logout_request + "</soapenv:Body>"
            "</soapenv:Envelope>"
        )

        # setup mocks
        self.validate_sign_patcher = patch.object(OneLogin_Saml2_Utils, "validate_sign")
        self.validate_sign_mock = self.validate_sign_patcher.start()

        self.uuid_patcher = patch("onelogin.saml2.utils.uuid4")
        self.uuid_mock = self.uuid_patcher.start()
        self.uuid_mock.hex = "80dd245883b84bd98dacbf3978af3d03"

        self.addCleanup(patch.stopall)

    @freeze_time("2020-04-09T08:31:46Z")
    def test_logout_request(self):
        response = self.client.post(
            self.url, data=self.logout_request_soap, content_type="text/xml"
        )

        self.assertEqual(response.status_code, 200)

        soap_tree = etree.fromstring(response.content)
        self.assertEqual(
            soap_tree.tag, "{http://schemas.xmlsoap.org/soap/envelope/}Envelope"
        )

        tree = OneLogin_Saml2_XML.remove_soap_envelope(response.content)
        self.assertEqual(
            tree.attrib,
            {
                "ID": "ONELOGIN_5ba93c9db0cff93f52b521d7420e43f6eda2784f",
                "Version": "2.0",
                "IssueInstant": "2020-04-09T08:31:46Z",
                "Destination": "https://preprod1.digid.nl/saml/idp/request_logout",
                "InResponseTo": "_1331125262",
            },
        )

        # Make sure Signature properties are as expected.
        signature = tree.xpath(
            "//ds:Signature", namespaces=OneLogin_Saml2_Constants.NSMAP
        )[0]
        signature.xpath(
            "//ds:SignatureValue", namespaces=OneLogin_Saml2_Constants.NSMAP
        )[0].text = ""
        signature.xpath("//ds:DigestValue", namespaces=OneLogin_Saml2_Constants.NSMAP)[
            0
        ].text = ""
        signature.xpath(
            "//ds:X509Certificate", namespaces=OneLogin_Saml2_Constants.NSMAP
        )[0].text = ""
        expected_signature = (
            '<ds:Signature xmlns:ds="http://www.w3.org/2000/09/xmldsig#" '
            'xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol" '
            'xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion" '
            'xmlns:soap="http://schemas.xmlsoap.org/soap/envelope/">'
            "<ds:SignedInfo>"
            '<ds:CanonicalizationMethod Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#"/>'
            '<ds:SignatureMethod Algorithm="http://www.w3.org/2001/04/xmldsig-more#rsa-sha256"/>'
            '<ds:Reference URI="#ONELOGIN_5ba93c9db0cff93f52b521d7420e43f6eda2784f">'
            "<ds:Transforms>"
            '<ds:Transform Algorithm="http://www.w3.org/2000/09/xmldsig#enveloped-signature"/>'
            '<ds:Transform Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#"/>'
            "</ds:Transforms>"
            '<ds:DigestMethod Algorithm="http://www.w3.org/2001/04/xmlenc#sha256"/>'
            "<ds:DigestValue></ds:DigestValue>"
            "</ds:Reference>"
            "</ds:SignedInfo>"
            "<ds:SignatureValue></ds:SignatureValue>"
            "<ds:KeyInfo>"
            "<ds:X509Data>"
            "<ds:X509Certificate></ds:X509Certificate>"
            "</ds:X509Data>"
            "</ds:KeyInfo>"
            "</ds:Signature>"
        )
        self.assertXMLEqual(
            etree.tostring(signature, pretty_print=True, encoding="unicode"),
            etree.tostring(
                etree.fromstring(expected_signature),
                pretty_print=True,
                encoding="unicode",
            ),
        )

        # check issuer
        issuer = tree.xpath("saml:Issuer", namespaces=OneLogin_Saml2_Constants.NSMAP)[0]
        self.assertEqual(issuer.text, "sp.example.nl/digid")

        # check status
        status_code = tree.xpath(
            "samlp:Status/samlp:StatusCode", namespaces=OneLogin_Saml2_Constants.NSMAP
        )[0]
        self.assertEqual(
            status_code.attrib["Value"], "urn:oasis:names:tc:SAML:2.0:status:Success"
        )

        # check session
        self.assertFalse("_auth_user_id" in self.client.session)

    def test_logout_empty_request(self):
        with self.assertLogs("digid_eherkenning.saml2", level="ERROR") as log_watcher:
            response = self.client.post(self.url, content_type="text/xml")

        self.assertEqual(response.status_code, 500)

        logs = [r.getMessage() for r in log_watcher.records]
        self.assertIn(
            "Logout request from Digid failed: SAML LogoutRequest body not found.", logs
        )

        expected_response = (
            '<soap:Envelope xmlns:soap="http://schemas.xmlsoap.org/soap/envelope/">'
            "<soap:Body>"
            "<soap:Fault>"
            "<faultcode>SOAP-ENV:Client</faultcode>"
            "<faultstring>SAML LogoutRequest body not found.</faultstring>"
            "</soap:Fault>"
            "</soap:Body>"
            "</soap:Envelope>"
        )
        self.assertEqual(
            response.content.decode().replace("\n", "").strip(), expected_response
        )

        # check session
        self.assertTrue("_auth_user_id" in self.client.session)

    def test_logout_incorrect_issuer(self):
        root_element = etree.fromstring(self.logout_request_soap)
        issuer = get_saml_element(root_element, ".//saml:Issuer")
        issuer.text = "Other issuer"

        with self.assertLogs("digid_eherkenning.saml2", level="ERROR") as log_watcher:
            response = self.client.post(
                self.url,
                data=etree.tostring(root_element, encoding="unicode"),
                content_type="text/xml",
            )

        self.assertEqual(response.status_code, 200)

        # check logs
        logs = [r.getMessage() for r in log_watcher.records]
        self.assertIn(
            "Logout request from Digid failed: Invalid issuer in the Logout Request "
            "(expected https://was-preprod1.digid.nl/saml/idp/metadata, got Other issuer)",
            logs,
        )

        # check status
        tree = OneLogin_Saml2_XML.remove_soap_envelope(response.content)
        status_code = tree.xpath(
            "samlp:Status/samlp:StatusCode", namespaces=OneLogin_Saml2_Constants.NSMAP
        )[0]
        self.assertEqual(
            status_code.attrib["Value"], "urn:oasis:names:tc:SAML:2.0:status:Responder"
        )

        # check session
        self.assertTrue("_auth_user_id" in self.client.session)

    def test_logout_no_name_id(self):
        root_element = etree.fromstring(self.logout_request_soap)
        name_id = get_saml_element(root_element, ".//saml:NameID")
        name_id.getparent().remove(name_id)

        with self.assertLogs("digid_eherkenning.saml2", level="ERROR") as log_watcher:
            response = self.client.post(
                self.url,
                data=etree.tostring(root_element, encoding="unicode"),
                content_type="text/xml",
            )

        self.assertEqual(response.status_code, 200)

        # check logs
        logs = [r.getMessage() for r in log_watcher.records]
        self.assertIn(
            "Logout request from Digid failed: NameID not found in the Logout Request",
            logs,
        )

        # check status
        tree = OneLogin_Saml2_XML.remove_soap_envelope(response.content)
        status_code = tree.xpath(
            "samlp:Status/samlp:StatusCode", namespaces=OneLogin_Saml2_Constants.NSMAP
        )[0]
        self.assertEqual(
            status_code.attrib["Value"], "urn:oasis:names:tc:SAML:2.0:status:Responder"
        )

        # check session
        self.assertTrue("_auth_user_id" in self.client.session)

    def test_logout_signature_incorrect_ref(self):
        root_element = etree.fromstring(self.logout_request_soap)
        signature_ref = root_element.find(
            ".//{%s}Reference" % OneLogin_Saml2_Constants.NS_DS
        )
        signature_ref.set("URI", "12345")

        with self.assertLogs("digid_eherkenning.saml2", level="ERROR") as log_watcher:
            response = self.client.post(
                self.url,
                data=etree.tostring(root_element, encoding="unicode"),
                content_type="text/xml",
            )

        self.assertEqual(response.status_code, 200)

        # check logs
        logs = [r.getMessage() for r in log_watcher.records]
        self.assertIn(
            "Logout request from Digid failed: Found an invalid Signed Element. Rejected",
            logs,
        )

        # check status
        tree = OneLogin_Saml2_XML.remove_soap_envelope(response.content)
        status_code = tree.xpath(
            "samlp:Status/samlp:StatusCode", namespaces=OneLogin_Saml2_Constants.NSMAP
        )[0]
        self.assertEqual(
            status_code.attrib["Value"], "urn:oasis:names:tc:SAML:2.0:status:Responder"
        )

        # check session
        self.assertTrue("_auth_user_id" in self.client.session)



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class DigidEherkenningConfig(AppConfig):
    name = "digid_eherkenning"
    verbose_name = _("DigiD, eHerkenning & eIDAS")
    default_auto_field = "django.db.models.AutoField"



from argparse import Action


class BooleanOptionalAction(Action):  # pragma: no cover
    """
    this class is copied from python3.9  for compatibility with python 3.7

    It should be deleted after python 3.7 support ends.
    """

    def __init__(
        self,
        option_strings,
        dest,
        default=None,
        type=None,
        choices=None,
        required=False,
        help=None,
        metavar=None,
    ):
        _option_strings = []
        for option_string in option_strings:
            _option_strings.append(option_string)

            if option_string.startswith("--"):
                option_string = "--no-" + option_string[2:]
                _option_strings.append(option_string)

        if help is not None and default is not None:
            help += " (default: %(default)s)"

        super().__init__(
            option_strings=_option_strings,
            dest=dest,
            nargs=0,
            default=default,
            type=type,
            choices=choices,
            required=required,
            help=help,
            metavar=metavar,
        )

    def __call__(self, parser, namespace, values, option_string=None):
        if option_string in self.option_strings:
            setattr(namespace, self.dest, not option_string.startswith("--no-"))

    def format_usage(self):
        return " | ".join(self.option_strings)
